@@ -1,24 +1,26 @@
 /*
 
-  Portions copyright (c) Nils McCarthy
-  Portions copyright (c) MainNerve, Inc.
-  Portions copyright (c) Ugen Antsilevitch
-  Portions copyright (c) Genuity, Inc. 
-
-
   This file is part of LFT.
 
   The LFT software provided in this Distribution is
-  Copyright 2002 MainNerve, Inc.
+  Copyright 2005 VOSTROM Holdings, Inc.
 
   The full text of our legal notices is contained in the file called
   COPYING, included with this Distribution.
- 
-  Authors: 
 
-         -  MainNerve Staff <software@mainnerve.com> 
-         -  Ugen Antsilevitch <ugen@xonix.com>
-         -  Nils McCarthy <nils@shkoo.com> 
+  Authors:
+
+  -  Victor Oppleman <lft@oppleman.com>
+  -  Eugene Antsilevitch <ugen@xonix.com>
+
+  Other copyrights and Former Authors:
+
+  -  Portions copyright (c) Victor Oppleman
+  -  Portions copyright (c) MainNerve, Inc.
+  -  Portions copyright (c) Eugene Antsilevitch
+  -  Portions copyright (c) Genuity, Inc. 
+  -  Portions copyright (c) Nils McCarthy
+
 */
 
 
@@ -76,6 +78,12 @@ typedef signed long n_time;
 #ifdef BSD
 #include <machine/limits.h>
 #endif
+#ifdef BSD_IP_STACK
+#include <errno.h>
+#include <sys/ioctl.h>
+#include <net/bpf.h>
+#include <pcap-int.h>
+#endif
 
 #ifdef sun
 #include <limits.h>
@@ -97,19 +105,20 @@ typedef signed long n_time;
 
 static char *hostname = NULL;
 static char *hostname_lsrr[9];
+static char *version = "2.31 (1/2005)";    /* set version string */
 static int hostname_lsrr_size = 0;
 static int dflag = 0;
 static int hop_info_size = 0;
 static int hop_info_length = 0;
 
 static int seq_start = 0; 
-static int dport = 80;
-static int sport = 53;		/* now we use tcp/53 dns-xfer */
+static int dport = 80;          /* set default to tcp/80 HTTP */
+static int sport = 53;		/* set default to tcp/53 dns-xfer */
 
 static int win_len = 32768;
 
 static int timeout_ms = 1000;	/* timeout between retries */
-static int retry_max = 3;	/* number of retries before giving up */
+static int retry_max = 2;	/* number of retries before giving up */
 static int retry_min = 1;	/* minimum number of checks per hop */
 static double scatter_ms = 20;	/* milleseconds between sends */
 static int ahead_limit = 5;	/* number of probes we can send
@@ -123,7 +132,7 @@ static int nostatus = 0;
 static int userdevsel = 0;
 static int resolve_names = 1;
 static int timetrace = 0;
-static int max_net_dev_input = 32;
+static int max_net_dev_input = 64;
 static int smart = 0;
 static int do_netlookup = 0;
 static int do_aslookup = 0;
@@ -138,6 +147,7 @@ int skip_header_len;
 
 #ifdef __CYGWIN__
 int recv_sock;
+static int		sock = -1;
 #endif
 
 unsigned short ip_id;
@@ -321,33 +331,59 @@ get_address(char *host)
 
   h = gethostbyname (host);
   if (!h) {
-    fprintf (stderr, "Unknown host: %s\n", host);
+    fprintf (stderr, "LFT: Unknown host: %s\n", host);
     exit (1);
   }
   addr = (struct in_addr *) h->h_addr;
   return addr->s_addr;
 }
 
+#ifdef __CYGWIN__
+char *
+lft_getifforremote(const char *remote)
+{
+  struct hostent *h;
+	struct sockaddr_in in, out;
+	unsigned long nBytesReturned;
+
+	/* Only do this once of course */
+	if (sock < 0) {
+		if ((sock = socket (AF_INET, SOCK_DGRAM, 0)) < 0) {
+			perror ("socket");
+			return NULL;
+		}
+	}
+
+	/* Ask what i/f  the packet should go out on */
+	in.sin_family = AF_INET;
+	in.sin_port = 0;
+	in.sin_addr.s_addr = get_address((char *)remote);
+	if (SOCKET_ERROR != WSAIoctl(sock, SIO_ROUTING_INTERFACE_QUERY, &in, sizeof(in), &out, sizeof(out), &nBytesReturned, NULL, NULL))
+		return lft_getifname(out.sin_addr);
+
+	/* not found */
+	return NULL;
+}
+#endif
+
 static void
 init_address (char *remote, char *pcap_dev)
 {
 
   struct hostent *h;
 
-  local_address.s_addr = lft_getifaddr (pcap_dev);	/* LJD */
+  local_address.s_addr = lft_getifaddr (pcap_dev);	
 
   h = gethostbyname (remote);
   if (!h) {
-    fprintf (stderr, "Unknown host: %s\n", remote);
+    fprintf (stderr, "LFT: Unknown host: %s\n", remote);
     exit (1);
   }
   remote_address = *(struct in_addr *) h->h_addr;
-  if (noisy && !nostatus) {
+  if (noisy) {
     print_host (local_address);
-    printf (":%d", sport);
+    printf (":%d\n", sport);
   }
-  if (!nostatus)
-    printf ("\n");
 }
 
 static void
@@ -374,7 +410,7 @@ open_sockets ()
   recv_sock = socket (PF_INET, SOCK_RAW, IPPROTO_IP);
   
   if (recv_sock < 0) {
-    perror ("raw socket");
+    perror ("LFT: raw socket");
     exit (1);
   }
   local_bind.sin_addr = local_address;
@@ -391,13 +427,13 @@ open_sockets ()
   send_sock = socket (PF_INET, SOCK_RAW, IPPROTO_RAW);
 #endif
   if (send_sock < 0) {
-    perror ("raw socket");
+    perror ("LFT: raw socket");
     exit (1);
   }
 #ifdef IP_HDRINCL
   if (setsockopt (send_sock, IPPROTO_IP, IP_HDRINCL, (char *) &on,
 		  sizeof (on)) < 0) {
-    perror ("IP_HDRINCL");
+    perror ("LFT: IP_HDRINCL");
     exit (1);
   }
 #endif
@@ -406,7 +442,7 @@ open_sockets ()
   local_bind.sin_port = 0;
   local_bind.sin_family = AF_INET;
   if (bind (send_sock, (struct sockaddr *)&local_bind, sizeof (local_bind)) < 0) {
-    perror ("bind");
+    perror ("LFT: bind");
     exit (1);
   }
 #endif
@@ -512,8 +548,8 @@ send_packet (short nhop, unsigned short ttl, unsigned int seq, unsigned char fla
   if (noisy) 
     printf("SEND TTL=%d TSEQ=%u FLAGS=%x\n", nhop, tseq, flags);
   else {
-    if (!nostatus)
-      printf("_");
+  /*  if (!nostatus)
+      printf(">"); */     
   }
 
   ts_last_sent = now;
@@ -648,7 +684,7 @@ finish ()
 
   gettimeofday (&trace_done_time, NULL);
 
-  if (noisy) printf ("Concluding with %d hops.\n", (num_hops+1));
+  if (noisy >= 2) printf ("Concluding with %d hops.\n", (num_hops+1));
   if (num_hops) {
     maxhop = num_hops;
     /* if we got a bunch of hits for the end host, display them all */
@@ -659,14 +695,14 @@ finish ()
     maxhop = hop_info_length - 1;
   }
 
-  if (num_hops && !nostatus)
-    printf (".\n\nTTL  LFT trace to ");
-  else {
-    if (!nostatus)
+  if (num_hops && !nostatus && !noisy)
+    printf ("|\n\nTTL  LFT trace to ");
+  else if (num_hops && !nostatus)
+    printf ("\nTTL  LFT trace to ");
+  else if (!nostatus)
       printf ("\n\nTTL  LFT trace to ");
-    if (nostatus)
+  else if (nostatus)
       printf ("\nTTL  LFT trace to ");
-  };
 
   print_host (remote_address);
   printf(":%d/tcp\n",dport);
@@ -690,8 +726,8 @@ finish ()
       printf("**   [firewall] the next gateway may statefully inspect packets\n");
     }
 
-    if ((hop_info[hopno].flags & HF_ENDPOINT) && (noreply >= ((maxhop - ttl_min)/2))) {
-      printf("**   [4.2-3 BSD bug] the next gateway may errantly reply with reused TTLs\n");
+    if ((hop_info[hopno].flags & HF_ENDPOINT) && (noreply >= ((maxhop - ttl_min)/2)) && num_hops > 3) {
+      printf("**   [4.2-3 BSD bug?] the next gateway may errantly reply with reused TTLs\n");
     }
 
 
@@ -762,10 +798,10 @@ finish ()
     printf("**   [%d/tcp failed]  Try alternate options or use -V to see packets.\n", dport);
   if (timetrace) {
     gettimeofday (&now, NULL);
-    printf ("\nLFT\'s trace took %.2f seconds.  ", 
+    printf ("\nLFT trace took %.2f seconds.  ", 
 	    (timediff_ms(begin_time, trace_done_time) / 1000));
     if (resolve_names || do_aslookup || do_netlookup) 
-      printf("Resolution required %.2f seconds.", (timediff_ms(trace_done_time, now) / 1000));
+      printf("Resolution took %.2f seconds.", (timediff_ms(trace_done_time, now) / 1000));
     printf("\n");
   }
   printf("\n"); 
@@ -828,8 +864,8 @@ check_timeouts ()
   }
   if (no_reply >= ahead_limit) {	/* we timed out. */
     if ((last_return + 3) * 2 < hop_info_length) {
-      printf("\n\nLFT can\'t seem to round-trip.  Local packet filter in the way?");
-      if (noisy == 2) printf("Will finish ONE\n");
+      if (need_reply < 3) printf("\n\nLFT can\'t reliably round-trip.  Local packet filter in the way?\n");
+      /* if (noisy == 2) printf("Will finish ONE\n"); */
       finish ();
     }
     /* otherwise search out for weird ttl's on return packets */
@@ -845,7 +881,7 @@ check_timeouts ()
     }
 
     if (num_hops > 0 && hop_info_length >= num_hops) {
-      if (noisy == 2) printf("I have blanks to fill -  RETURN \n");
+      if (noisy == 2) printf("I have blanks to fill in the path, returning\n");
       return;			/* we know how long the path is --
 				 * wait to fill in the blanks */
     }
@@ -853,7 +889,7 @@ check_timeouts ()
     nhop = hop_info_length++;
     send_hop(nhop);
   } else { 
-    if (noisy) printf("Everyone responded.  Moving on...\n");
+    if (noisy >= 2) printf("Everyone responded.\n");
     for (nhop = ttl_min; nhop < hop_info_length; nhop++) {
       if (hop_info[nhop].num_sent < retry_min && hop_info[nhop].num_sent <= retry_max /*&& hop_info[nhop].ts_last_recv.tv_sec*/) {
 	send_hop(nhop);
@@ -861,7 +897,7 @@ check_timeouts ()
       }
     }
 
-    if (noisy) printf("Will finish TWO\n");
+    /* if (noisy == 2) printf("Will finish TWO\n"); */
     finish ();
   } 
 }
@@ -881,7 +917,7 @@ recv_packet (unsigned int seq, struct in_addr ipaddr, int icmp_type)
 
   if (!tp) {
     if (noisy)
-      printf("Not for us?\n");
+      printf("Packet not meant for us?\n");
     else
       if (!nostatus) printf("?");
 	
@@ -890,16 +926,16 @@ recv_packet (unsigned int seq, struct in_addr ipaddr, int icmp_type)
 
   if (tp->recv.tv_sec) {
     if (noisy)
-      printf ("duplicate?\n");
+      printf ("Duplicate packet?\n");
     else
       if (!nostatus) printf("!");
     return;
   }
 
   if (noisy) {
-    printf ("RECV RSEQ=%u\n", seq);
+    printf ("RCVD RSEQ=%u\n", seq);
   } else {
-    if (!nostatus) printf("_");
+    if (!nostatus) printf(".");
   }
 
   tp->recv = now;
@@ -920,7 +956,7 @@ recv_packet (unsigned int seq, struct in_addr ipaddr, int icmp_type)
 			      * tcp packet in return - meaning destination */
       	num_hops = tp->hopno;
       	tp->is_done = 1;
-       	if (noisy) printf ("Looks like we made it.\n");
+       	if (noisy >= 2) printf ("Looks like we made it.\n");
       }
     }
 
@@ -961,7 +997,7 @@ void process_packet (const u_char *packet)
       	return; * not for us */
 
 
-    if (noisy) printf ("ICMP SEQ=%u received\n", ntohl (tcp->th_seq));
+    if (noisy) printf ("ICMP RCVD SEQ=%u\n", ntohl (tcp->th_seq));
     recv_packet (ntohl (tcp->th_seq) , orig_ip->ip_src,
 		 (icmp->icmp_type == ICMP_TIMXCEED) ? -2 : icmp->icmp_code);
     break;
@@ -985,7 +1021,7 @@ void process_packet (const u_char *packet)
     if (tcp->th_flags & TH_SYN)
       if (noisy) printf ("SYN ");
     if (noisy) printf (" SEQ=%lu ACK=%lu ", ntohl (tcp->th_seq), ntohl (tcp->th_ack));
-    if (noisy) printf ("from %s\n", inet_ntoa (ip->ip_src));
+    if (noisy) printf ("FROM %s\n", inet_ntoa (ip->ip_src));
 
     /*if (ntohl(tcp->th_ack) < seq_start || ntohl(tcp->th_ack) > seq_start + trace_packet_info_length + 1)
     	return; * not for us */
@@ -1010,7 +1046,7 @@ static void
 usage (char *prog)
 {
   fprintf (stderr,
-	   "\nLFT - the alternative traceroute tool for network (reverse) engineers\n\nUsage: %s [<options>] [<gateway> <...>] <hostname:dport>\n"
+	   "\nLayer Four Traceroute (LFT)\n\n    - the alternative traceroute tool for network (reverse) engineers\n\nUsage: %s [<options>] [<gateway> <...>] <target:destport>\n"
 	   "\nOptions are:\n"
 	   "  -d <dport>       destination port number\n"
 	   "  -s <sport>       source port number\n"
@@ -1021,7 +1057,7 @@ usage (char *prog)
 	   "  -t <timeout ms>  maximum RTT before assuming packet was dropped\n"
 	   "  -l <min ttl>     minimum TTL to use on outgoing packets\n"
 	   "  -q <sequence>    set the initial sequence number (ISN)\n"
-           "  -D <device/ip>   network device (name or IP address) to use (e.g., \"en1\")\n"
+           "  -D <device|ip>   network device name or IP address (e.g., \"en1\" or \"1.2.3.4\")\n"
            "  -H <ttl>         maximum number of hops to traverse (max TTL of packets)\n"
            "  -i               disable \"stop on ICMP\" other than TTL expired\n"
            "  -n               disable use of the DNS resolver to display hostnames\n"
@@ -1031,7 +1067,7 @@ usage (char *prog)
            "  -T               enable display of LFT's execution timer\n"
            "  -S               suppress the status bar (only show the completed trace)\n"
            "  -V               print a lot of debugging garbage including packets\n"
-           "  -E               enable LFT's stateful Engine to detect firewalls\n"
+           "  -E/-e            enable LFT's stateful Engine to detect firewalls\n"
 	   "  -v               show version information\n"  "\n"
 	   "Default is: %s -d %d -m %d -M %d -a %d -c %.0f -t %d -H %d -s %d\n\n",
 	   prog, prog, dport, retry_min, retry_max, ahead_limit,
@@ -1040,12 +1076,15 @@ usage (char *prog)
 }
 
 static void
-version (void)
+show_version (void)
 {
   fprintf (stderr, "\n"
-	   "LFT version 2.2 2003/05/07 compiled for " HOST_SYSTEM_TYPE "\n\n"
+	   "LFT version %s compiled for " HOST_SYSTEM_TYPE "\n\n"
 	   "    Compile-time options:\n\n"
 #if defined(BSD_IP_STACK)
+	   "      + BSDSTACK\n"
+#endif
+#if defined(BSD)
 	   "      + BSD\n"
 #endif
 #if defined(linux)
@@ -1067,7 +1106,7 @@ version (void)
 #else
 	   "      + no IP_HDRINCL\n"
 #endif
-	   "\n");
+	   "\n", version);
   exit (0);
 }
 
@@ -1111,10 +1150,11 @@ main (int argc, char **argv)
   int use_fins = 0;
   char *cp;
   struct timeval tb;
+  unsigned int bpfimmflag;
 
   setbuf (stdout, NULL);
 
-  while ((ch = getopt (argc, argv, "EANc:FVq:d:M:l:H:m:a:s:t:D:p:vinSTw:")) != EOF)
+  while ((ch = getopt (argc, argv, "EeANc:FVq:d:M:l:H:m:a:s:t:D:p:vinSTw:")) != EOF)
     switch (ch) {
     case 'F':
       if (smart) {
@@ -1165,6 +1205,14 @@ main (int argc, char **argv)
       }
       smart = 1;
       break;
+    case 'e':
+      if (use_fins) {
+        fprintf (stderr,
+                 "\nLFT warning: TCP flags are selected automatically using (-e) option.\n             Ignoring smart Engine (-e) option and using FINs-only (-F).\n\n");
+        break;
+      }
+      smart = 1;
+      break;
     case 'S':
       nostatus = 1;
       break;
@@ -1172,7 +1220,7 @@ main (int argc, char **argv)
       {
 	if (strlen (optarg) > max_net_dev_input) {
 	  fprintf (stderr,
-		   "Net interface names are limited to %d characters. (e.g., \"eth0\" or \"ppp0\")\n",
+		   "Net interface names are limited to %d characters. (e.g., \"eth0\" or \"ppp0\" or \"10.10.10.10\")\n",
 		   max_net_dev_input);
 	  exit (1);
 	}
@@ -1206,10 +1254,11 @@ main (int argc, char **argv)
       break_on_icmp = 0;
       break;
     case 'v':
-      version ();
+      show_version ();
       break;
     case 'V':
       noisy++;
+      nostatus = 1; 
       break;
     default:
       usage (argv[0]);
@@ -1219,6 +1268,9 @@ main (int argc, char **argv)
     usage (argv[0]);
 
 
+  if (noisy)
+      printf ("Layer Four Traceroute (LFT) version %s\n", version);
+
   gettimeofday (&tb, NULL);
   /* eventually this might want to use /dev/urandom or
    * something on machines that have it.  otherwise,
@@ -1231,11 +1283,6 @@ main (int argc, char **argv)
    */
   srand(tb.tv_sec ^ (tb.tv_usec * 2147));
 
-  if (!seq_start) {
-    seq_start = rand();
-    if (noisy) printf("Setting my ISN = %d\n", seq_start);
-  }
-
   hostname = argv[optind++];
   hostname_lsrr_size = 0;
   while (optind < argc) {
@@ -1269,11 +1316,11 @@ main (int argc, char **argv)
   /* if not given network interface, select one automatically */
   if (!userdevsel) {
 #ifdef __CYGWIN__
-    pcap_dev = "none";
+    pcap_dev = lft_getifforremote(hostname);
 #else
     pcap_dev = pcap_lookupdev (ebuf);
     if (pcap_dev == NULL) {
-      fprintf (stderr, "%s\n", ebuf);
+      fprintf (stderr, "LFT: %s\n", ebuf);
       exit (1);
     };
 #endif
@@ -1283,7 +1330,7 @@ main (int argc, char **argv)
       /* specified by ip address -- look up device. */
       pcap_dev = lft_getifname(addr);
       if (!pcap_dev) {
-	fprintf (stderr, "Could not find matching interface with IP address %s\n", userdev);
+	fprintf (stderr, "LFT: Could not find an interface with IP address %s\n", userdev);
 	exit(1);
       }
     } else
@@ -1292,7 +1339,6 @@ main (int argc, char **argv)
 
 #ifdef __CYGWIN__
   skip_header_len = 0;
-  pcap_dev = "eth0";
 #else
   /* now, fix pcap's linux implementation */
   if (!strncmp (pcap_dev, "lo:", 3) || !strcmp (pcap_dev, "lo")) {
@@ -1303,16 +1349,28 @@ main (int argc, char **argv)
   else				/* assume ethernet */
     skip_header_len = sizeof (struct ether_header);
 #endif
-  if (noisy && !nostatus)
+  if (noisy)
     printf ("Using device %s, ", pcap_dev);
 #ifndef __CYGWIN__
   pd = pcap_open_live (pcap_dev, 1600, 0, 20, ebuf);
   if (!pd) {
-    fprintf (stderr, "%s\n", ebuf);
+    fprintf (stderr, "\nSorry, that interface isn\'t available to LFT.  Try another or fix:\nERROR: %s\n", ebuf);
     exit (1);
   }
+#ifdef BSD_IP_STACK
+  bpfimmflag = 1;
+  if ( ioctl(pd->fd, BIOCIMMEDIATE, &bpfimmflag) < 0) {
+    fprintf(stderr, "BIOCIMMEDIATE: %s",
+    pcap_strerror(errno));
+  }
 #endif
+#endif
+
   init_address (hostname, pcap_dev);
+  if (!seq_start) {
+     seq_start = rand();
+     if (noisy) printf("Setting my ISN = %d\n", seq_start);
+  }
   open_sockets ();
 
   if (do_netlookup || do_aslookup)
